@@ -1,8 +1,21 @@
-from distributed import config  # isort:skip; load distributed configuration first
-from distributed import widgets  # isort:skip; load distributed widgets second
+from __future__ import annotations
+
+# isort: off
+from distributed import config  # load distributed configuration first
+from distributed import widgets  # load distributed widgets second
+
+# isort: on
 
 import atexit
+import weakref
 
+# This finalizer registers an atexit handler that has to happen before
+# distributed registers its handlers, otherwise we observe hangs on
+# cluster shutdown when using the UCX comms backend. See
+# https://github.com/dask/distributed/issues/7726 for more discussion
+# of the problem and the search for long term solutions
+
+weakref.finalize(lambda: None, lambda: None)
 import dask
 from dask.config import config  # type: ignore
 
@@ -22,10 +35,18 @@ from distributed.client import (
     wait,
 )
 from distributed.core import Status, connect, rpc
-from distributed.deploy import Adaptive, LocalCluster, SpecCluster, SSHCluster
+from distributed.deploy import (
+    Adaptive,
+    LocalCluster,
+    SpecCluster,
+    SSHCluster,
+    SubprocessCluster,
+)
 from distributed.diagnostics.plugin import (
+    CondaInstall,
     Environ,
     NannyPlugin,
+    PackageInstall,
     PipInstall,
     SchedulerPlugin,
     UploadDirectory,
@@ -42,6 +63,7 @@ from distributed.queues import Queue
 from distributed.scheduler import KilledWorker, Scheduler
 from distributed.security import Security
 from distributed.semaphore import Semaphore
+from distributed.spans import span
 from distributed.threadpoolexecutor import rejoin
 from distributed.utils import CancelledError, TimeoutError, sync
 from distributed.variable import Variable
@@ -94,3 +116,69 @@ def _():
     """
     global _python_shutting_down
     _python_shutting_down = True
+
+
+__all__ = [
+    "Actor",
+    "ActorFuture",
+    "Adaptive",
+    "BaseActorFuture",
+    "CancelledError",
+    "Client",
+    "CompatibleExecutor",
+    "CondaInstall",
+    "Environ",
+    "Event",
+    "Future",
+    "KilledWorker",
+    "LocalCluster",
+    "Lock",
+    "MultiLock",
+    "Nanny",
+    "NannyPlugin",
+    "PackageInstall",
+    "PipInstall",
+    "Pub",
+    "Queue",
+    "Reschedule",
+    "SSHCluster",
+    "Scheduler",
+    "SchedulerPlugin",
+    "Security",
+    "Semaphore",
+    "SpecCluster",
+    "Status",
+    "Sub",
+    "SubprocessCluster",
+    "TimeoutError",
+    "UploadDirectory",
+    "UploadFile",
+    "Variable",
+    "Worker",
+    "WorkerPlugin",
+    "as_completed",
+    "config",
+    "connect",
+    "dask",
+    "default_client",
+    "fire_and_forget",
+    "futures_of",
+    "get_client",
+    "get_task_metadata",
+    "get_task_stream",
+    "get_versions",
+    "get_worker",
+    "local_client",
+    "performance_report",
+    "print",
+    "progress",
+    "rejoin",
+    "rpc",
+    "secede",
+    "span",
+    "sync",
+    "wait",
+    "warn",
+    "widgets",
+    "worker_client",
+]
