@@ -2,8 +2,10 @@ from __future__ import annotations
 
 import asyncio
 import logging
+import warnings
 from collections import defaultdict
 from timeit import default_timer
+from typing import ClassVar
 
 from tlz import groupby, valmap
 
@@ -102,7 +104,9 @@ class Progress(SchedulerPlugin):
         logger.debug("Set up Progress keys")
 
         for k in errors:
-            self.transition(k, None, "erred", exception=True)
+            self.transition(
+                k, None, "erred", stimulus_id="progress-setup", exception=True
+            )
 
     def transition(self, key, start, finish, *args, **kwargs):
         if key in self.keys and start == "processing" and finish == "memory":
@@ -139,9 +143,19 @@ class Progress(SchedulerPlugin):
 class MultiProgress(Progress):
     """Progress variant that keeps track of different groups of keys
 
-    See Progress for most details.  This only adds a function ``func=``
-    that splits keys.  This defaults to ``key_split`` which aligns with naming
-    conventions chosen in the dask project (tuples, hyphens, etc..)
+    See Progress for most details.
+
+    Parameters
+    ----------
+
+    func : Callable (deprecated)
+        Function that splits keys. This defaults to ``key_split`` which
+        aligns with naming conventions chosen in the dask project (tuples,
+        hyphens, etc..)
+
+    group_by : Callable | Literal["spans"] | Literal["prefix"], default: "prefix"
+        How to group keys to display multiple bars. Defaults to "prefix",
+        which uses ``key_split`` from dask project
 
     State
     -----
@@ -160,10 +174,24 @@ class MultiProgress(Progress):
     """
 
     def __init__(
-        self, keys, scheduler=None, func=key_split, minimum=0, dt=0.1, complete=False
+        self,
+        keys,
+        scheduler=None,
+        *,
+        func=None,
+        group_by="prefix",
+        minimum=0,
+        dt=0.1,
+        complete=False,
     ):
-        self.func = func
-        name = f"multi-progress-{tokenize(keys, func, minimum, dt, complete)}"
+        if func is not None:
+            warnings.warn(
+                "`func` is deprecated, use `group_by`", category=DeprecationWarning
+            )
+            group_by = func
+        self.group_by = key_split if group_by in (None, "prefix") else group_by
+        self.func = None
+        name = f"multi-progress-{tokenize(keys, group_by, minimum, dt, complete)}"
         super().__init__(
             keys, scheduler, minimum=minimum, dt=dt, complete=complete, name=name
         )
@@ -190,6 +218,22 @@ class MultiProgress(Progress):
         if not self.keys:
             self.stop(exception=None, key=None)
 
+        if self.group_by == "spans":
+            spans_ext = self.scheduler.extensions["spans"]
+            span_defs = spans_ext.spans if spans_ext else None
+
+            def group_key(k):
+                span_id = self.scheduler.tasks[k].group.span_id
+                span_name = ", ".join(span_defs[span_id].name) if span_defs else span_id
+                return span_name, span_id
+
+            group_keys = {k: group_key(k) for k in self.all_keys}
+            self.func = group_keys.get
+        elif self.group_by == "prefix":
+            self.func = key_split
+        else:
+            self.func = self.group_by
+
         # Group keys by func name
         self.keys = valmap(set, groupby(self.func, self.keys))
         self.all_keys = valmap(set, groupby(self.func, self.all_keys))
@@ -198,7 +242,9 @@ class MultiProgress(Progress):
                 self.keys[k] = set()
 
         for k in errors:
-            self.transition(k, None, "erred", exception=True)
+            self.transition(
+                k, None, "erred", stimulus_id="multiprogress-setup", exception=True
+            )
         logger.debug("Set up Progress keys")
 
     def transition(self, key, start, finish, *args, **kwargs):
@@ -298,7 +344,10 @@ class AllProgress(SchedulerPlugin):
 class GroupTiming(SchedulerPlugin):
     """Keep track of high-level timing information for task group progress"""
 
-    name = "group-timing"
+    name: ClassVar[str] = "group-timing"
+    time: list[float]
+    compute: dict[str, list[float]]
+    nthreads: list[float]
 
     def __init__(self, scheduler):
         self.scheduler = scheduler
@@ -309,17 +358,17 @@ class GroupTiming(SchedulerPlugin):
         # Initialize our data structures.
         self._init()
 
-    def _init(self):
+    def _init(self) -> None:
         """Shared initializatoin code between __init__ and restart"""
         now = time()
 
         # Timestamps for tracking compute durations by task group.
         # Start with length 2 so that we always can compute a valid dt later.
-        self.time: list[float] = [now] * 2
+        self.time = [now] * 2
         # The amount of compute since the last timestamp
-        self.compute: dict[str, list[float]] = {}
+        self.compute = {}
         # The number of threads at the time
-        self.nthreads: list[float] = [self.scheduler.total_nthreads] * 2
+        self.nthreads = [self.scheduler.total_nthreads] * 2
 
     def transition(self, key, start, finish, *args, **kwargs):
         # We are mostly interested in when tasks complete for now, so just look
