@@ -1,22 +1,30 @@
+from __future__ import annotations
+
 import logging
 
 import msgpack
 
+import dask.config
+
+from distributed.protocol import pickle
 from distributed.protocol.compression import decompress, maybe_compress
 from distributed.protocol.serialize import (
+    Pickled,
     Serialize,
     Serialized,
+    ToPickle,
     merge_and_deserialize,
     msgpack_decode_default,
     msgpack_encode_default,
     serialize_and_split,
 )
 from distributed.protocol.utils import msgpack_opts
+from distributed.utils import ensure_memoryview
 
 logger = logging.getLogger(__name__)
 
 
-def dumps(
+def dumps(  # type: ignore[no-untyped-def]
     msg, serializers=None, on_error="message", context=None, frame_split_size=None
 ) -> list:
     """Transform Python message to bytestream suitable for communication
@@ -45,31 +53,56 @@ def dumps(
 
             header["compression"] = tuple(compression)
 
+        def create_serialized_sub_frames(obj: Serialized | Serialize) -> list:
+            if isinstance(obj, Serialized):
+                sub_header, sub_frames = obj.header, obj.frames
+            else:
+                sub_header, sub_frames = serialize_and_split(
+                    obj,
+                    serializers=serializers,
+                    on_error=on_error,
+                    context=context,
+                    size=frame_split_size,
+                )
+                _inplace_compress_frames(sub_header, sub_frames)
+            sub_header["num-sub-frames"] = len(sub_frames)
+            sub_header = msgpack.dumps(
+                sub_header, default=msgpack_encode_default, use_bin_type=True
+            )
+            return [sub_header] + sub_frames
+
+        def create_pickled_sub_frames(obj: Pickled | ToPickle) -> list:
+            if isinstance(obj, Pickled):
+                sub_header, sub_frames = obj.header, obj.frames
+            else:
+                sub_frames = []
+                sub_header = {
+                    "pickled-obj": pickle.dumps(
+                        obj.data,
+                        # In to support len() and slicing, we convert `PickleBuffer`
+                        # objects to memoryviews of bytes.
+                        buffer_callback=lambda x: sub_frames.append(
+                            ensure_memoryview(x)
+                        ),
+                    )
+                }
+                _inplace_compress_frames(sub_header, sub_frames)
+
+            sub_header["num-sub-frames"] = len(sub_frames)
+            sub_header = msgpack.dumps(sub_header)
+            return [sub_header] + sub_frames
+
         frames = [None]
 
         def _encode_default(obj):
-            typ = type(obj)
-            if typ is Serialize or typ is Serialized:
+            if isinstance(obj, (Serialize, Serialized)):
                 offset = len(frames)
-                if typ is Serialized:
-                    sub_header, sub_frames = obj.header, obj.frames
-                else:
-                    sub_header, sub_frames = serialize_and_split(
-                        obj,
-                        serializers=serializers,
-                        on_error=on_error,
-                        context=context,
-                        size=frame_split_size,
-                    )
-                    _inplace_compress_frames(sub_header, sub_frames)
-                sub_header["num-sub-frames"] = len(sub_frames)
-                frames.append(
-                    msgpack.dumps(
-                        sub_header, default=msgpack_encode_default, use_bin_type=True
-                    )
-                )
-                frames.extend(sub_frames)
+                frames.extend(create_serialized_sub_frames(obj))
                 return {"__Serialized__": offset}
+            elif isinstance(obj, (ToPickle, Pickled)):
+                offset = len(frames)
+                frames.extend(create_pickled_sub_frames(obj))
+                return {"__Pickled__": offset}
             else:
                 return msgpack_encode_default(obj)
 
@@ -84,6 +117,8 @@ def dumps(
 def loads(frames, deserialize=True, deserializers=None):
     """Transform bytestream back into Python value"""
 
+    allow_pickle = dask.config.get("distributed.scheduler.pickle")
+
     try:
 
         def _decode_default(obj):
@@ -105,8 +140,20 @@ def loads(frames, deserialize=True, deserializers=None):
                     )
                 else:
                     return Serialized(sub_header, sub_frames)
-            else:
-                return msgpack_decode_default(obj)
+
+            offset = obj.get("__Pickled__", 0)
+            if offset > 0:
+                sub_header = msgpack.loads(frames[offset])
+                offset += 1
+                sub_frames = frames[offset : offset + sub_header["num-sub-frames"]]
+                if allow_pickle:
+                    return pickle.loads(sub_header["pickled-obj"], buffers=sub_frames)
+                else:
+                    raise ValueError(
+                        "Unpickle on the Scheduler isn't allowed, set `distributed.scheduler.pickle=true`"
+                    )
+
+            return msgpack_decode_default(obj)
 
         return msgpack.loads(
             frames[0], object_hook=_decode_default, use_list=False, **msgpack_opts
